from setuptools import setup

setup(
    name='frametests',
    packages=['frametests'],
    version='0.1.1',
    description='A library for testing and analyzing data integrity.',
    author='Atai Ambus',
    license='MIT',
)
